/*
See LICENSE folder for this sample’s licensing information.

Abstract:
Class containing static methods for halftone descreening.
*/

import Accelerate
import simd
import UIKit

class HalftoneDescreener {
    // MARK: 2D FFT to demo halftone descreening
    static func descreen(sourceImage: UIImage, halftoneSample: UIImage, threshold: Float) -> UIImage? {
        assert(sourceImage.size == CGSize(width: 1024, height: 1024) &&
            halftoneSample.size == CGSize(width: 1024, height: 1024),
               "Image must be 1024 x 1024")
        
        let width = Int(sourceImage.size.width)
        let height = Int(sourceImage.size.height)
        let pixelCount = width * height
        let n = pixelCount / 2
        
        var sourceImageReal = [Float](repeating: 0, count: n)
        var sourceImageImaginary = [Float](repeating: 0, count: n)
        
        var halftoneSampleReal = [Float](repeating: 0, count: n)
        var halftoneSampleImag = [Float](repeating: 0, count: n)
        
        let result: UIImage? = sourceImageReal.withUnsafeMutableBufferPointer { sourceImageRealPtr in
            sourceImageImaginary.withUnsafeMutableBufferPointer { sourceImageImaginaryPtr in
                halftoneSampleReal.withUnsafeMutableBufferPointer { halftoneSampleRealPtr in
                    halftoneSampleImag.withUnsafeMutableBufferPointer { halftoneSampleImagPtr in
                        
                        var sourceImageSplitComplex = DSPSplitComplex(realp: sourceImageRealPtr.baseAddress!,
                                                                      imagp: sourceImageImaginaryPtr.baseAddress!)
                        
                        imageToComplex(sourceImage,
                                       splitComplex: &sourceImageSplitComplex)
                        
                        var halftoneSampleSplitComplex = DSPSplitComplex(realp: halftoneSampleRealPtr.baseAddress!,
                                                                         imagp: halftoneSampleImagPtr.baseAddress!)
                        
                        imageToComplex(halftoneSample,
                                       splitComplex: &halftoneSampleSplitComplex)
                        
                        return fft_2D(sourceImageSplitComplex: &sourceImageSplitComplex,
                                      halftoneSampleSplitComplex: &halftoneSampleSplitComplex,
                                      size: sourceImage.size,
                                      threshold: threshold)
                    }
                }
            }
        }
        
        return result
    }
    
    static let fftSetUp = vDSP.FFT2D(width: 1024,
                                     height: 1024,
                                     ofType: DSPSplitComplex.self)
    
    static func fft_2D(sourceImageSplitComplex: inout DSPSplitComplex,
                       halftoneSampleSplitComplex: inout DSPSplitComplex,
                       size: CGSize,
                       threshold: Float) -> UIImage? {
                
        let width = Int(size.width)
        let height = Int(size.height)
        let pixelCount = width * height
        let n = pixelCount / 2
        
        var sourceImage_floatPixelsReal_frequency = [Float](repeating: 0,
                                                          count: n)
        var sourceImage_floatPixelsImag_frequency = [Float](repeating: 0,
                                                            count: n)
        
        var halftoneSample_floatPixelsReal_frequency = [Float](repeating: 0,
                                                               count: n)
        var halftoneSample_floatPixelsImag_frequency = [Float](repeating: 0,
                                                               count: n)
        var halftoneSampleAmplitude = [Float](repeating: 0,
                                                count: n)
        
        sourceImage_floatPixelsReal_frequency.withUnsafeMutableBufferPointer { sourceRealPtr in
            sourceImage_floatPixelsImag_frequency.withUnsafeMutableBufferPointer { sourceImagPtr in
                halftoneSample_floatPixelsReal_frequency.withUnsafeMutableBufferPointer { halftoneRealPtr in
                    halftoneSample_floatPixelsImag_frequency.withUnsafeMutableBufferPointer { halftoneImagPtr in
                        
                        var sourceImage_floatPixels_frequency = DSPSplitComplex(
                            realp: sourceRealPtr.baseAddress!,
                            imagp: sourceImagPtr.baseAddress!)
                        
                        fftSetUp?.transform(input: sourceImageSplitComplex,
                                            output: &sourceImage_floatPixels_frequency,
                                            direction: .forward)
                                            
                        var halftoneSample_floatPixels_frequency = DSPSplitComplex(
                            realp: halftoneRealPtr.baseAddress!,
                            imagp: halftoneImagPtr.baseAddress!)
                        
                        fftSetUp?.transform(input: halftoneSampleSplitComplex,
                                            output: &halftoneSample_floatPixels_frequency,
                                            direction: .forward)
                        
                        vDSP.squareMagnitudes(halftoneSample_floatPixels_frequency,
                                              result: &halftoneSampleAmplitude)
                    }
                }
            }
        }
   
        // all values in `halftoneSampleAmplitude` greater than `threshold` to -1, else +1...
        let outputConstant: Float = -1
        
        vDSP.threshold(halftoneSampleAmplitude,
                       to: threshold,
                       with: .signedConstant(outputConstant),
                       result: &halftoneSampleAmplitude)
        
        // all negative values in `halftoneSampleAmplitude` to 0...
        vDSP.clip(halftoneSampleAmplitude,
                  to: 0 ... 1,
                  result: &halftoneSampleAmplitude)
        
        // multiply frequency domain pixels by values in `halftoneSampleAmplitude`...
        halftoneSampleAmplitude[0] = 1
       
        var floatPixelsReal_spatial = [Float](repeating: 0,
                                              count: n)
        var floatPixelsImag_spatial = [Float](repeating: 0,
                                              count: n)
        
        let image: UIImage? = sourceImage_floatPixelsReal_frequency.withUnsafeMutableBufferPointer { sourceRealPtr in
            sourceImage_floatPixelsImag_frequency.withUnsafeMutableBufferPointer { sourceImagPtr in
                floatPixelsReal_spatial.withUnsafeMutableBufferPointer { destinationRealPtr in
                    floatPixelsImag_spatial.withUnsafeMutableBufferPointer { destinationImagPtr in
                        
                        var sourceImage_floatPixels_frequency = DSPSplitComplex(
                            realp: sourceRealPtr.baseAddress!,
                            imagp: sourceImagPtr.baseAddress!)
                        
                        vDSP.multiply(sourceImage_floatPixels_frequency,
                                      by: halftoneSampleAmplitude,
                                      result: &sourceImage_floatPixels_frequency)
                        
                        // perform inverse FFT to create image from frequency domain data...
                        
                        var floatPixels_spatial = DSPSplitComplex(realp: destinationRealPtr.baseAddress!,
                                                                  imagp: destinationImagPtr.baseAddress!)
                        
                        fftSetUp?.transform(input: sourceImage_floatPixels_frequency,
                                            output: &floatPixels_spatial,
                                            direction: .inverse)
                        
                        return imageFromPixelSource(&floatPixels_spatial,
                                                    width: width,
                                                    height: height,
                                                    bitmapInfo: CGBitmapInfo(rawValue: 0))
                    }
                }
            }
        }
        
        return image
    }
    
    static func imageToComplex(_ image: UIImage, splitComplex splitComplexOut: inout DSPSplitComplex) {
        guard let cgImage = image.cgImage else {
            fatalError("unable to generate cgimage")
        }
        
        let pixelCount = Int(image.size.width * image.size.height)
        
        let pixelData = cgImage.dataProvider?.data
        
        let pixelsArray = Array(UnsafeBufferPointer(start: CFDataGetBytePtr(pixelData),
                                                    count: pixelCount))
        
        let floatPixels = vDSP.integerToFloatingPoint(pixelsArray,
                                                      floatingPointType: Float.self)
        
        let interleavedPixels = stride(from: 1, to: floatPixels.count, by: 2).map {
            return DSPComplex(real: floatPixels[$0.advanced(by: -1)],
                              imag: floatPixels[$0])
        }
        
        vDSP.convert(interleavedComplexVector: interleavedPixels,
                     toSplitComplexVector: &splitComplexOut)
    }
    
    static func imageFromPixelSource(_ pixelSource: inout DSPSplitComplex,
                                     width: Int, height: Int,
                                     denominator: Float = pow(2, 29),
                                     bitmapInfo: CGBitmapInfo) -> UIImage? {
        
        let pixelCount = width * height
        let n = vDSP_Length(pixelCount / 2)
        let stride = vDSP_Stride(1)
        
        // multiply all float values (1 / denominator) * 255
        let multiplier = [Float](repeating: Float((1 / denominator) * 255),
                                 count: pixelCount / 2)

        vDSP.multiply(pixelSource,
                      by: multiplier,
                      result: &pixelSource)
        
        // Clip values to 0...255
        var low: Float = 0
        var high: Float = 255

        vDSP_vclip(pixelSource.realp,
                   stride,
                   &low,
                   &high,
                   pixelSource.realp,
                   stride,
                   n)
        
        vDSP_vclip(pixelSource.imagp,
                   stride,
                   &low,
                   &high,
                   pixelSource.imagp,
                   stride, n)
        
        var uIntPixels = [UInt8](repeating: 0,
                                 count: pixelCount)

        let floatPixels = [Float](fromSplitComplex: pixelSource,
                                  scale: 1,
                                  count: pixelCount)
        
        vDSP.convertElements(of: floatPixels,
                             to: &uIntPixels,
                             rounding: .towardZero)
        
        
        let result: UIImage? = uIntPixels.withUnsafeMutableBufferPointer { uIntPixelsPtr in
            
            let buffer = vImage_Buffer(data: uIntPixelsPtr.baseAddress!,
                                       height: vImagePixelCount(height),
                                       width: vImagePixelCount(width),
                                       rowBytes: width)
            
            if
                let format = vImage_CGImageFormat(bitsPerComponent: 8,
                                                  bitsPerPixel: 8,
                                                  colorSpace: CGColorSpaceCreateDeviceGray(),
                                                  bitmapInfo: bitmapInfo),
                let cgImage = try? buffer.createCGImage(format: format) {
                
                return UIImage(cgImage: cgImage)
            } else {
                print("Unable to create `CGImage`.")
                return nil
            }
        }

        
        return result
    }
}
